import typing
from abc import ABC, abstractmethod

from sqlalchemy import Boolean, Column, func
from sqlalchemy.sql.elements import BooleanClauseList
from sqlalchemy.sql.expression import ColumnElement

from souswift_core.filters.base import Comparison, Filter


class Equal(Comparison):
    def compare(
        self, attr: typing.Union[Column, typing.Any], target: typing.Any
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return attr == target


class NotEqual(Comparison):
    def compare(
        self, attr: typing.Union[Column, typing.Any], target: typing.Any
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return attr != target


class Greater(Comparison):
    def compare(
        self, attr: typing.Union[Column, typing.Any], target: typing.Any
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return attr > target


class GreaterEqual(Comparison):
    def compare(
        self, attr: typing.Union[Column, typing.Any], target: typing.Any
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return attr >= target


class Lesser(Comparison):
    def compare(
        self, attr: typing.Union[Column, typing.Any], target: typing.Any
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return attr < target


class LesserEqual(Comparison):
    def compare(
        self, attr: typing.Union[Column, typing.Any], target: typing.Any
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return attr <= target


class Like(Comparison):
    def compare(
        self, attr: typing.Union[Column, typing.Any], target: typing.Any
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return attr.like(f'%{target}%')


class InsensitiveLike(Comparison):
    def compare(
        self, attr: typing.Union[Column, typing.Any], target: typing.Any
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return attr.ilike(f'%{target}%')


class Contains(Comparison):
    def compare(
        self, attr: typing.Union[Column, typing.Any], target: typing.Any
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return attr.in_(target)


class Excludes(Comparison):
    def compare(
        self, attr: typing.Union[Column, typing.Any], target: typing.Any
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return attr.not_in(target)


class Null(Comparison):
    def compare(
        self, attr: typing.Union[Column, typing.Any], isnull: bool
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return attr.is_(None) if isnull else attr.is_not(None)


class JSONContains(Comparison):
    def compare(
        self, attr: typing.Union[Column, typing.Any], target: typing.Any
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return func.json_contains(attr, f'"{target}"')  # type: ignore


class EmptyJson(Comparison):
    def compare(
        self, attr: typing.Union[Column, typing.Any], target: typing.Any
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return func.json_length(attr) == 0


class RelatedComp(ABC):
    def __init__(self, where: Filter):
        self._where = where

    def __bool__(self):
        return self._where.__bool__()

    @abstractmethod
    def compare(
        self, attr: type, target: typing.Union[Column, typing.Any]
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        ...


class RelatedWhere(RelatedComp):
    def compare(
        self, attr: type, target: typing.Union[Column, typing.Any]
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return self._where.where(attr)


class RelatedHas(RelatedComp):
    def compare(
        self, attr: type, target: typing.Union[Column, typing.Any]
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return attr.has(self._where.where(attr))


class RelatedAny(RelatedComp):
    def compare(
        self, attr: type, target: typing.Union[Column, typing.Any]
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return attr.any(self._where.where(attr))


class RelatedEmpty(RelatedComp):
    def compare(
        self, attr: type, target: typing.Union[Column, typing.Any]
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        return ~target.any()



import operator
from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional, Union

from sqlalchemy import Column, and_, false, func, or_, true
from sqlalchemy.orm.relationships import RelationshipProperty
from sqlalchemy.sql.elements import BooleanClauseList
from src.database.base import Entity


class Filter:
    def where(self, entity: type[Entity]):
        pass


class CompareOptions(str, Enum):
    EQUAL = 'eq'
    NOT_EQUAL = 'ne'
    GREATER = 'gt'
    GREATER_EQUAL = 'ge'
    LESSER = 'lt'
    LESSER_EQUAL = 'le'
    LIKE = 'like'
    ILIKE = 'ilike'
    JSON_CONTAINS = 'contains'
    EMPTY = 'empty'

    def in_exclude(self):
        return self in [
            CompareOptions.LIKE,
            CompareOptions.ILIKE,
            CompareOptions.JSON_CONTAINS,
            CompareOptions.EMPTY,
        ]


@dataclass
class FieldFilter(Filter):
    field: str
    value: Optional[Any]
    related: str = ''
    comparison: CompareOptions = CompareOptions.EQUAL

    def __post_init__(self):
        if isinstance(self.value, bool):
            self.value = true() if self.value else false()

    def where(self, entity: type[Entity]):
        if self.related:
            return self.related_attr(entity)
        return self.attr(entity)

    def related_attr(self, entity: type[Entity]):
        related_entity = self._attr(entity, self.related)
        return self.attr(related_entity.entity.class_)  # type: ignore

    def attr(self, entity: type[Entity]):
        attr = self._attr(entity, self.field)
        if self.value is None:
            return True
        if not self.comparison.in_exclude():
            return getattr(operator, self.comparison.value)(attr, self.value)
        if self.comparison == CompareOptions.JSON_CONTAINS:
            return func.json_contains(attr, f'"{self.value}"') == 1
        if self.comparison == CompareOptions.EMPTY:
            return func.json_length(attr) == 0
        if self.comparison == CompareOptions.LIKE:
            return attr.like(f'%{self.value}%')
        return attr.ilike(f'%{self.value}%')

    @staticmethod
    def _attr(
        entity: type[Entity], field: str
    ) -> Union[Column, RelationshipProperty]:
        result = getattr(entity, field, None)
        if not result:
            raise NotImplementedError
        return result


@dataclass(init=False)
class FilterJoins(Filter):
    operator: type[BooleanClauseList]
    filters: tuple[Filter, ...]

    def __init__(self, *filters: Filter) -> None:
        self.filters = filters

    def where(self, entity: type[Entity]):
        return self.operator(*(f.where(entity) for f in self.filters))


class OrFilter(FilterJoins):
    @property
    def operator(self):
        return or_


class AndFilter(FilterJoins):
    @property
    def operator(self):
        return and_



from contextlib import asynccontextmanager
from typing import AsyncGenerator

from context_handler import AsyncContext
from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncConnection

from souswift_core.providers.database.base import DatabaseProvider


class ConnectionProvider(DatabaseProvider):
    @asynccontextmanager
    async def acquire(self) -> AsyncGenerator[AsyncConnection, None]:
        async with self.engine.begin() as connection:
            yield connection

    @staticmethod
    def is_closed(client: AsyncConnection):
        return client.closed

    @staticmethod
    async def close_client(client: AsyncConnection):
        await client.close()

    async def health_check(self):
        async with self.acquire() as connection:
            await connection.execute(text('SELECT 1'))
        self.logger.info('Database Connection succeeded')

    Context = AsyncContext[AsyncConnection]
